# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
import sys
from logging import WARNING, getLevelName
from traceback import format_exc, format_exception
from typing import Iterator, List
from unittest import TestCase

from six import text_type

import filters as f
from filters.base import ExceptionHandler


class ExceptionHandlerTestCase(TestCase):
    def setUp(self):
        super(ExceptionHandlerTestCase, self).setUp()

        self.handler = ExceptionHandler()

    def test_invalid_value(self):
        """
        Sends an invalid value to the handler.
        """
        message = 'Needs more cowbell.'
        context = {
            'key':      'test',
            'value':    "(Don't Fear) The Reaper",
        }

        # When ExceptionHandler encounters an invalid value, it raises
        # an exception.
        # The exception always has the same type (FilterError) so that
        # the caller can capture it.
        with self.assertRaises(f.FilterError) as exc:
            self.handler.handle_invalid_value(message, False, context)

        self.assertEqual(text_type(exc.exception), message)
        self.assertEqual(exc.exception.context, context)

    def test_exception(self):
        """
        Sends an exception to the handler.
        """
        message = 'An exception occurred!'
        context = {
            'key':      'test',
            'value':    "(Don't Fear) The Reaper",
        }

        #
        # ExceptionHandler converts any exception it encounters into a
        # FilterError.
        # The exception always has the same type (FilterError) so that
        # the caller can capture it.
        #
        # Note that the FilterError completely replaces the original
        # exception, but it leaves the traceback intact.
        # To make things more convenient, Filters add the exception
        # info to the context dict, but you can still use
        # `sys.exc_info()[2]` to inspect the original exception's
        # stack.
        #
        # :see: importer.core.f.BaseFilter._invalid_value
        #
        try:
            # Raise an exception so that the handler has a traceback to
            # work with.
            # Note that the ValueError's message will get replaced (but
            # it can still be accessed via the traceback).
            exc = ValueError('Needs more cowbell.')
            exc.context = context
            raise exc
        except ValueError as e:
            with self.assertRaises(f.FilterError) as ar_context:
                self.handler.handle_exception(message, e)

            self.assertEqual(text_type(ar_context.exception), message)
            self.assertEqual(ar_context.exception.context, context)


class MemoryLogHandler(logging.Handler):
    """
    A log handler that retains all of its records in a list in memory,
    so that we can test :py:class:`LogHandler`.

    This class is similar in function (though not in purpose) to
    BufferingHandler.

    References:
      - :py:class:`logging.handlers.BufferingHandler`
    """
    def __init__(self, level=logging.NOTSET):
        # type: (int) -> None
        super(MemoryLogHandler, self).__init__(level)

        self._records           = [] # type: List[logging.LogRecord]
        self.max_level_emitted  = logging.NOTSET

    def __getitem__(self, index):
        # type: (int) -> logging.LogRecord
        """
        Returns the log message at the specified index.
        """
        return self._records[index]

    def __iter__(self):
        # type: () -> Iterator[logging.LogRecord]
        """
        Creates an iterator for the collected records.
        """
        return iter(self._records)

    def __len__(self):
        # type: () -> int
        """Returns the number of log records collected."""
        return len(self._records)

    @property
    def records(self):
        # type: () -> List[logging.LogRecord]
        """
        Returns all log messages that the handler has collected.
        """
        return self._records[:]

    def clear(self):
        """
        Removes all log messages that this handler has collected.
        """
        del self._records[:]

    def emit(self, record):
        # type: (logging.LogRecord) -> None
        """
        Records the log message.
        """
        # Remove `exc_info` to reclaim memory.
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = ''.join(format_exception(*record.exc_info))

            record.exc_info = None

        self._records.append(record)
        self.max_level_emitted = max(self.max_level_emitted, record.levelno)


class LogHandlerTestCase(TestCase):
    def setUp(self):
        super(LogHandlerTestCase, self).setUp()

        self.logs = MemoryLogHandler()

        logger = logging.getLogger(__name__)
        logger.addHandler(self.logs)
        logger.setLevel(logging.DEBUG)

        self.handler = f.LogHandler(logger, WARNING)

    def test_invalid_value(self):
        """
        Sends an invalid value to the handler.
        """
        message = 'Needs more cowbell.'
        context = {
            'key':      'test',
            'value':    "(Don't Fear) The Reaper",
        }

        self.handler.handle_invalid_value(message, False, context)

        self.assertEqual(len(self.logs.records), 1)
        self.assertEqual(self.logs[0].msg, message)
        self.assertEqual(getattr(self.logs[0], 'context'), context)

        # The log message level is set in the handler's initializer.
        self.assertEqual(self.logs[0].levelname, getLevelName(WARNING))

        # No exception info for invalid values (by default).
        self.assertIsNone(self.logs[0].exc_text)

    def test_exception(self):
        """
        Sends an exception to the handler.
        """
        message = 'An exception occurred!'
        context = {
            'key':      'test',
            'value':    "(Don't Fear) The Reaper",
        }

        try:
            # Raise an exception so that the handler has a traceback to
            # work with.
            # Note that the ValueError's message will get replaced (but
            # it can still be accessed via the traceback).
            exc = ValueError('Needs more cowbell.')
            exc.context = context
            raise exc
        except ValueError as e:
            original_traceback = format_exc()

            self.handler.handle_exception(message, e)

            self.assertEqual(len(self.logs.records), 1)

            self.assertEqual(len(self.logs.records), 1)
            self.assertEqual(self.logs[0].msg, message)
            self.assertEqual(getattr(self.logs[0], 'context'), context)

            # The log message level is set in the handler's
            # initializer.
            # Note that both invalid values and exceptions have the
            # same log level.
            self.assertEqual(self.logs[0].levelname, getLevelName(WARNING))

            # Traceback is captured for exceptions.
            self.assertEqual(self.logs[0].exc_text, original_traceback)


class MemoryHandlerTestCase(TestCase):
    def setUp(self):
        super(MemoryHandlerTestCase, self).setUp()

        self.handler = f.MemoryHandler()

    def test_invalid_value(self):
        """
        Sends an invalid value to the handler.
        """
        code    = 'foo'
        key     = 'test'
        message = 'Needs more cowbell.'
        context = {
            'code':     code,
            'key':      key,
            'value':    "(Don't Fear) The Reaper",
        }

        self.handler.handle_invalid_value(message, False, context)

        # Add a couple of additional messages, just to make sure the
        # handler stores them correctly.
        self.handler.handle_invalid_value(
            message     = 'Test message 1',
            exc_info    = False,
            context     = {'key':  key},
        )
        self.handler.handle_invalid_value('Test message 2', False, {})

        # As filter messages are captured, they are sorted according to
        # their contexts' ``key`` values.
        # If a message doesn't have a ``key`` value, an empty string is
        # used.
        self.assertListEqual(sorted(self.handler.messages.keys()), ['', key])

        filter_message_0 = self.handler.messages[key][0]
        self.assertIsInstance(filter_message_0, f.FilterMessage)
        self.assertEqual(filter_message_0.code, code)
        self.assertEqual(filter_message_0.message, message)
        self.assertEqual(filter_message_0.context, context)
        self.assertIsNone(filter_message_0.exc_info)

        filter_message_1 = self.handler.messages[key][1]
        self.assertIsInstance(filter_message_1, f.FilterMessage)
        self.assertEqual(filter_message_1.code, 'Test message 1')
        self.assertEqual(filter_message_1.message, 'Test message 1')
        self.assertEqual(filter_message_1.context, {'key': key})
        self.assertIsNone(filter_message_1.exc_info)

        filter_message_blank = self.handler.messages[''][0]
        self.assertIsInstance(filter_message_blank, f.FilterMessage)
        self.assertEqual(filter_message_blank.code, 'Test message 2')
        self.assertEqual(filter_message_blank.message, 'Test message 2')
        self.assertEqual(filter_message_blank.context, {})
        self.assertIsNone(filter_message_blank.exc_info)

        self.assertFalse(self.handler.has_exceptions)
        self.assertListEqual(self.handler.exc_info, [])

    def test_exception(self):
        """
        Sends an exception to the handler.
        """
        code    = 'error'
        key     = 'test'
        message = 'An exception occurred!'
        context = {
            'code':     code,
            'key':      key,
            'value':    "(Don't Fear) The Reaper",
        }

        try:
            # Raise an exception so that the handler has a traceback to
            # work with.
            # Note that the ValueError's message will get replaced (but
            # it can still be accessed via the traceback).
            exc = ValueError('Needs more cowbell.')
            exc.context = context
            raise exc
        except ValueError as e:
            original_traceback = format_exc()

            self.handler.handle_exception(message, e)

            self.assertListEqual(list(self.handler.messages.keys()), [key])

            filter_message_0 = self.handler.messages[key][0]
            self.assertIsInstance(filter_message_0, f.FilterMessage)
            self.assertEqual(filter_message_0.code, code)
            self.assertEqual(filter_message_0.message, message)
            self.assertEqual(filter_message_0.context, context)

            # Exception traceback is captured automatically.
            self.assertEqual(filter_message_0.exc_info, original_traceback)

            self.assertTrue(self.handler.has_exceptions)

            # By default, the handler does NOT keep the traceback
            # object.
            # :see: test_capture_exc_info
            self.assertListEqual(self.handler.exc_info, [])

    def test_capture_exc_info(self):
        """
        The handler is configured to capture :py:func:`sys.exc_info` in
        the event of an exception.

        This is generally turned off because the filter already
        captures a formatted traceback in the event of an
        exception, so there is no need to store the raw traceback
        object.

        However, there are a few cases where it is useful to collect
        this, such as when you want to send exceptions to a logger
        that expects `sys.exc_info()`.
        """
        self.handler.capture_exc_info = True

        try:
            # Raise an exception so that the handler has a traceback to
            # work with.
            raise ValueError('I gotta have more cowbell, baby!')
        except ValueError as e:
            self.handler.handle_exception('An exception occurred!', e)

            self.assertTrue(self.handler.has_exceptions)
            self.assertListEqual(self.handler.exc_info, [sys.exc_info()])



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from inspect import getmembers as get_members, isabstract as is_abstract, \
    isclass as is_class, ismodule as is_module
from logging import getLogger
from typing import Any, Dict, Generator, Text, Tuple, Type, Union
from warnings import warn

from class_registry import EntryPointClassRegistry
from pkg_resources import EntryPoint, iter_entry_points

from filters.base import BaseFilter

__all__ = [
    'FilterExtensionRegistry',
    'GROUP_NAME',
]

GROUP_NAME = 'filters.extensions'
"""
The key to use when declaring entry points in your library's
``setup.py`` file.

Example::

   setup(
     ...
     entry_points = {
       'filters.extensions': [
         # Declare each filter with its own entry point.
         'Country=filters_iso:Country',
         'Currency=filters_iso:Currency',
       ],
     },
   )

Filters that are loaded this way are accessible from
:py:data:`filters.ext` (not imported into the global namespace because
it gives IDEs a heart attack).
"""

logger = getLogger(__name__)

legacy_warned = False
"""
Ensures the legacy extensions loader warning is only shown once.

References:
  - :py:func:`iter_filters_in`
"""

class FilterExtensionRegistry(EntryPointClassRegistry):
    """
    Creates a registry that can be used to dynamically load 3rd-party
    filters into the (nearly) top-level namespace.
    """
    def __init__(self, group=GROUP_NAME):
        super(FilterExtensionRegistry, self).__init__(group)

    def __getattr__(self, item):
        # type: (Text) -> Type[BaseFilter]
        return self[item]

    def __repr__(self):
        return repr(self._get_cache())

    def _get_cache(self):
        # type: () -> Dict[Text, Type[BaseFilter]]
        if self._cache is None:
            self._cache = {}

            try:
                for target in iter_entry_points(self.group): # type: EntryPoint
                    filter_ = target.load()

                    ift_result = is_filter_type(filter_)

                    if ift_result is True:
                        logger.debug(
                            'Registering extension filter '
                            '{cls.__module__}.{cls.__name__} as {name}.'.format(
                                cls     = filter_,
                                name    = target.name,
                            ),
                        )

                        self._cache[target.name] = filter_

                    else:
                        logger.debug(
                            'Using legacy extension loader for '
                            '{target.name} ({reason}).'.format(
                                reason  = ift_result,
                                target  = target,
                            ),
                        )

                        self._cache.update(iter_filters_in(filter_))
            except DeprecationWarning:
                # The user has ``simplefilter('error')`` set; reset the
                # cache so that the next time we try to load extension
                # filters, we don't miss anything.
                self._cache = None
                raise

        # noinspection PyTypeChecker
        return self._cache

    @staticmethod
    def create_instance(class_, *args, **kwargs):
        # type: (type, ...) -> Any
        if args or kwargs:
            return class_(*args, **kwargs)

        return class_



def is_filter_type(target):
    # type: (Any) -> Union[bool, Text]
    """
    Returns whether the specified object can be registered as a filter.

    :return:
        Returns ``True`` if the object is a filter.
        Otherwise, returns a string indicating why it is not valid.
    """
    if not is_class(target):
        return 'not a class'

    if not issubclass(target, BaseFilter):
        return 'does not extend BaseFilter'

    if is_abstract(target):
        return 'abstract class'

    return True


def iter_filters_in(target):
    # type: (Any) -> Generator[Tuple[Text, Type[BaseFilter]]]
    """
    Iterates over all filters in the specified module/class.
    """
    global legacy_warned
    if not legacy_warned:
        # Set the global flag to ``True`` first, in case the user has
        # ``simplefilter('error')`` set.
        legacy_warned = True

        warn(
            'Legacy extension loader is deprecated and will be removed in '
                'Filters v1.4.  '
                'See http://filters.readthedocs.io/en/latest/extensions.html#legacy-extensions-loader '
                'for more information.',

            DeprecationWarning,
        )

    ift_result = is_filter_type(target)

    if ift_result is True:
        logger.debug(
            'Registering extension filter '
            '{cls.__module__}.{cls.__name__}.'.format(
                cls = target,
            ),
        )

        yield target.__name__, target
    elif is_module(target):
        for member_name, member in get_members(target):
            member_ift_result = is_filter_type(member)

            if member_ift_result is True:
                logger.debug(
                    'Registering extension filter '
                    '{cls.__module__}.{cls.__name__}.'.format(
                        cls = member,
                    ),
                )

                yield member.__name__, member
            else:
                logger.debug(
                    'Ignoring {module}.{name} ({reason})'.format(
                        module  = target.__name__,
                        name    = member_name,
                        reason  = member_ift_result,
                    ),
                )
    elif is_class(target):
        logger.debug(
            'Ignoring {cls.__module__}.{cls.__name__} ({reason}).'.format(
                cls     = target,
                reason  = ift_result,
            ),
        )
    else:
        logger.debug(
            'Ignoring {target!r} ({reason}).'.format(
                reason  = ift_result,
                target  = target,
            ),
        )



# coding=utf-8
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from os.path import dirname
from unittest import TestCase
from warnings import catch_warnings, simplefilter

from pkg_resources import working_set

from filters.extensions import FilterExtensionRegistry
from filters.macros import FilterMacroType
from test import TestFilterAlpha, TestFilterBravo


def setUpModule():
    #
    # Install a fake distribution that we can use to inject entry
    # points at runtime.
    #
    # The side effects from this are pretty severe, but they (very
    # probably) only impact this test, and they are undone as soon as
    # the process terminates.
    #
    working_set.add_entry(dirname(__file__))


class FilterExtensionRegistryTestCase(TestCase):
    def test_happy_path(self):
        """
        Loading filters automatically via entry points.

        References:
          - /test/_support/filter_extension.egg-info/entry_points.txt
          - :py:func:`setUpModule`
        """
        # For this test, we will use a different entry point key, to
        # minimize the potential for side effects.
        registry = FilterExtensionRegistry('filters.extensions_test')

        # Note that the filters are registered using the entry point
        # names, which conventionally matches the class name (but it
        # doesn't strictly have to).
        alpha = registry.Alfred
        self.assertIs(alpha, TestFilterAlpha)

        # You can also instantiate filters using parameters.
        bravo = registry.Bruce(name='Batman')
        self.assertIsInstance(bravo, TestFilterBravo)
        self.assertEqual(bravo.name, 'Batman')

        # I couldn't find any Batman characters whose name begins with
        # C... and "Commissioner Gordon" doesn't count!
        charlie = registry.Catwoman
        # Note that :py:data:`test.TestFilterCharlie` is a filter
        # macro.
        self.assertTrue(issubclass(charlie, FilterMacroType))

        # Check that the correct number of extension filters were
        # registered.
        self.assertEqual(len(registry), 3)

    def test_legacy_extension_loader(self):
        """
        Loading filters using the legacy extension loader (for versions
        of filters < 1.3).

        References:
          - /test/_support/filter_extension_legacy.egg-info/entry_points.txt
          - :py:func:`setUpModule`
        """
        with catch_warnings():
            simplefilter('error', DeprecationWarning)

            # For this test, we will use a different entry point key,
            # to minimize the potential for side effects.
            registry = FilterExtensionRegistry('filters.extensions_test_legacy')

            # The legacy extensions loader will issue a warning the
            # first time it runs.
            with self.assertRaises(DeprecationWarning):
                dir(registry)

            # When using the legacy loader, the filter name must always
            # match the class name.
            alpha = registry.TestFilterAlpha
            self.assertIs(alpha, TestFilterAlpha)

            # You can also instantiate filters using parameters.
            bravo = registry.TestFilterBravo(name='Batman')
            self.assertIsInstance(bravo, TestFilterBravo)
            self.assertEqual(bravo.name, 'Batman')

            # I couldn't find any Batman characters whose name begins with
            # C... and "Commissioner Gordon" doesn't count!
            charlie = registry.TestFilterCharlie
            # Note that :py:data:`test.TestFilterCharlie` is a filter
            # macro.
            self.assertTrue(issubclass(charlie, FilterMacroType))

            # Check that the correct number of extension filters were
            # registered.
            self.assertEqual(len(registry), 3)



# coding=utf-8
from __future__ import absolute_import, unicode_literals

import json
from itertools import starmap
from pprint import pformat
from traceback import format_exception
from typing import Any, Callable, Mapping, Sequence
from unittest import TestCase

from collections import OrderedDict
from six import iterkeys, string_types

from filters.base import BaseFilter
from filters.handlers import FilterRunner

__all__ = [
    'BaseFilterTestCase',
]


def sorted_dict(value):
    # type: (Mapping) -> Any
    """
    Sorts a dict's keys to avoid leaking information about the
    backend's handling of unordered dicts.
    """
    if isinstance(value, Mapping):
        return OrderedDict(
            (key, sorted_dict(value[key]))
                for key in sorted(iterkeys(value))
        )

    elif isinstance(value, Sequence) and not isinstance(value, string_types):
        return list(map(sorted_dict, value))

    else:
        return value


class BaseFilterTestCase(TestCase):
    """
    Base functionality for request filter unit tests.

    Prevents typos from causing invalid test passes/failures by
    abstracting the Filter type out of filtering ops; just set
    ``filter_type`` at the top of your test case, and then use
    ``assertFilterPasses`` and ``assertFilterErrors`` to check use
    cases.
    """
    filter_type = None # type: Callable[[...], BaseFilter]

    class unmodified(object):
        """
        Used by ``assertFilterPasses`` so that you can omit the
        ``expected_value`` parameter.
        """
        pass

    class skip_value_check(object):
        """
        Used by ``assertFilterPasses`` to skip checking the filtered
        value.  This is useful for tests where a simple equality
        check is not practical.

        Note:  If you use ``skip_value_check``, you should add extra
        assertions to your test to make sure the filtered value
        conforms to expectations.
        """
        pass

    def assertFilterPasses(self, runner, expected_value=unmodified):
        """
        Asserts that the FilterRunner returns the specified value,
        without errors.

        :param runner:
            Usually a FilterRunner instance, but you can pass in the
            test value itself if you want (it will automatically be run
            through ``_filter``).

        :param expected_value:
            The expected value for ``runner.cleaned_data``.

            If omitted, the assertion will check that the incoming
            value is returned unmodified.
        """
        self.assertFilterErrors(runner, {}, expected_value)

    def assertFilterErrors(self, runner, expected_codes, expected_value=None):
        """
        Asserts that the FilterRunner generates the specified error
        codes.

        :param runner:
            Usually a FilterRunner instance, but you can pass in the
            test value itself if you want (it will automatically be
            run through `_filter`).

        :param expected_value:
            Expected value for ``runner.cleaned_data`` (usually
            ``None``).
        """
        if not isinstance(runner, FilterRunner):
            runner = self._filter(runner) # type: FilterRunner

        # First check to make sure no unhandled exceptions occurred.
        if runner.has_exceptions:
            # noinspection PyTypeChecker
            self.fail(
                'Unhandled exceptions occurred while filtering the '
                'request payload:\n\n{tracebacks}\n\n'
                'Filter Messages:\n\n{messages}'.format(
                    messages = pformat(dict(runner.filter_messages)),

                    tracebacks = pformat(list(
                        starmap(format_exception, runner.exc_info)
                    )),
                )
            )

        if isinstance(expected_codes, list):
            expected_codes = {'': expected_codes}

        if runner.error_codes != expected_codes:
            # noinspection PyTypeChecker
            self.fail(
                'Filter generated unexpected error codes (expected '
                '{expected}):\n\n{messages}'.format(
                    expected    = json.dumps(sorted_dict(expected_codes)),
                    messages    = pformat(dict(runner.filter_messages)),
                ),
            )

        check_value = (
                (self.skip_value_check is not True)
            and (expected_value is not self.skip_value_check)
        )

        if check_value:
            self._check_filter_value(
                runner.cleaned_data,
                runner.data
                    if expected_value is self.unmodified
                    else expected_value
            )

    def _filter(self, *args, **kwargs):
        # type: (...) -> FilterRunner
        """
        Applies the Filter to the specified value.

        Generally, you don't have to use this method in your test case,
        unless you want to specify Filter options.

        Example::

            self.filter_type = Min

            # Min().apply(42)
            self.assertFilterPasses(42)

            # Min(min_val=40).apply(42)
            self.assertFilterPasses(self._filter(42, min_val=40))

        :param args:
            Positional params to pass to the Filter's initializer.

        :param kwargs:
            Keyword params to pass to the Filter's initializer.
        """
        if not callable(self.filter_type):
            self.fail('{cls}.filter_type is not callable.'.format(
                cls = type(self).__name__,
            ))

        if not args:
            self.fail(
                'First argument to {cls}._filter '
                'must be the filtered value.'.format(
                    cls = type(self).__name__,
                ),
            )

        return FilterRunner(
            starting_filter     = self.filter_type(*args[1:], **kwargs),
            incoming_data       = args[0],
            capture_exc_info    = True,
        )

    def _check_filter_value(self, cleaned_data, expected):
        """
        Checks the value returned by the Filter, used by
        ``assertFilterPasses``.

        In certain cases, it may be useful to override this method in
        your test case subclass.

        :param cleaned_data:
            ``cleaned_data`` from the FilterRunner.
        """
        self.assertEqual(cleaned_data, expected)
